#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/stat.h>

#include "config.h"
#include "imapfilter.h"
#include "data.h"


extern char logfile[PATH_MAX];
extern unsigned int options;


/*
 * Find the path to configuration file, open it and call parse_config().
 */
int read_config(char *cfg)
{
    int r;
    FILE *fp;
    char *home = NULL;

    if (!cfg) {
	cfg = (char *) xmalloc(PATH_MAX * sizeof(char));

	home = getenv("HOME");

	snprintf(cfg, PATH_MAX, "%s/%s", home, ".imapfilterrc");
    }
#ifdef DEBUG
    printf("debug: configuration file: '%s'\n", cfg);
#endif
#ifdef CHECK_PERMISSIONS
    check_permissions(cfg);
#endif
    fp = fopen(cfg, "r");

    if (!fp)
	fatal(ERROR_FILE_OPEN, "imapfilter: opening config file %s; %s\n",
	      cfg, strerror(errno));

    if ((r = parse_config(fp)))
	fatal(ERROR_CONFIG_PARSE,
	      "imapfilter: parse error in config file at row %d\n", r);

    fclose(fp);

#ifdef DEBUG
    printf("debug: options: %0#10x\n", options);
#endif

    return 0;
}


#ifdef CHECK_PERMISSIONS
/*
 * Check the permissions of the configuration file.
 */
int check_permissions(char *cfg)
{
    struct stat fs;

    if (stat(cfg, &fs)) {
	error("imapfilter: getting file %s status; %s\n", cfg,
	      strerror(errno));
	return ERROR_TRIVIAL;
    }
    if (!S_ISREG(fs.st_mode)) {
	error("imapfilter: file %s not a regular file\n", cfg);
	return ERROR_TRIVIAL;
    }
    if ((fs.st_mode & 00777) != (S_IRUSR | S_IWUSR)) {
	error("imapfilter: warning: improper config file %s permissions\n"
	      "imapfilter: warning: file's mode should be 600 not %o\n",
	      cfg, fs.st_mode & 00777);
	return ERROR_TRIVIAL;
    }
    return 0;
}
#endif

/*
 * Parse configuration file.
 */
int parse_config(FILE * fp)
{
    int i, r = 0;
    unsigned int row = 0;
    char line[LINE_MAX];
    regex_t creg[12];
    regmatch_t match[7];
    const char *reg[12] = {
	"^([[:blank:]]*\n|#.*\n)$",
	"^[[:blank:]]*ACCOUNT[[:blank:]]+([[:graph:]]*):([[:graph:]]*)@([[:alnum:].-]+)(:[[:digit:]]+)?[[:blank:]]*\n$",
	"^[[:blank:]]*FOLDER[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+([[:graph:]]+)[[:blank:]]*\n$",
	"^[[:blank:]]*FILTER[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]*([[:blank:]]OR|[[:blank:]]AND)?[[:blank:]]*\n$",
	"^[[:blank:]]*ACTION[[:blank:]]+(DELETE|COPY[[:blank:]]+([[:graph:]]+)|MOVE[[:blank:]]+([[:graph:]]+)|LIST)[[:blank:]]*([[:graph:]]*)[[:blank:]]*\n$",
	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*(ANSWERED|DELETED|DRAFT|FLAGGED|NEW|OLD|RECENT|SEEN|UNANSWERED|UNDELETED|UNDRAFT|UNFLAGGED|UNSEEN)[[:blank:]]*\n$",
	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*(BCC|BODY|CC|FROM|SUBJECT|TEXT|TO)[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]*\n$",
	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*(HEADER)[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]*\n$",
	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*(LARGER|SMALLER)[[:blank:]]+([[:digit:]]+)[[:blank:]]*\n$",
	"^[[:blank:]]*JOB[[:blank:]]+([[:alnum:],_-]+)[[:blank:]]+([[:alnum:],_-]+)[[:blank:]]*\n$",
	"^[[:blank:]]*(LOGFILE)[[:blank:]]*=[[:blank:]]*([[:print:]]+)\n$",
	"^[[:blank:]]*(HEADERS)[[:blank:]]*=[[:blank:]]*(YES|NO)[[:blank:]]*\n$"
    };

    for (i = 0; i < 12; i++)
	regcomp(&creg[i], reg[i], REG_EXTENDED | REG_ICASE);

    while (fgets(line, LINE_MAX - 1, fp)) {
	row++;
	if (!regexec(&creg[0], line, 0, match, 0))
	    continue;
	else if (!regexec(&creg[1], line, 6, match, 0))
	    set_account(line, match);
	else if (!regexec(&creg[2], line, 3, match, 0))
	    r = set_mboxgrp(line, match);
	else if (!regexec(&creg[3], line, 3, match, 0))
	    r = set_filter(line, match);
	else if (!regexec(&creg[4], line, 5, match, 0))
	    r = set_action(line, match);
	else if (!regexec(&creg[5], line, 7, match, 0) ||
		 !regexec(&creg[6], line, 7, match, 0) ||
		 !regexec(&creg[7], line, 7, match, 0) ||
		 !regexec(&creg[8], line, 7, match, 0))
	    r = set_mask(line, match);
	else if (!regexec(&creg[9], line, 3, match, 0))
	    r = set_job(line, match);
	else if (!regexec(&creg[10], line, 3, match, 0) ||
		 !regexec(&creg[11], line, 3, match, 0))
	    set_options(line, match);
	else
	    return row;

	if (r == ERROR_CONFIG_PARSE)
	    return row;
    }

    for (i = 0; i < 12; i++)
	regfree(&creg[i]);

    destroy_data();

    return 0;
}


/*
 * Set other options found in config file.
 */
void set_options(char *line, regmatch_t * match)
{
    if (!strncasecmp(line + match[1].rm_so, "logfile", 7)) {
	if (!*logfile)
	    strncat(logfile, line + match[2].rm_so,
		    min((match[2].rm_eo - match[2].rm_so), PATH_MAX - 1));
    } else {
	if (!strncasecmp(line + match[2].rm_so, "yes", 3))
	    options |= OPTION_HEADERS;
	else if (options & OPTION_HEADERS)
	    options ^= OPTION_HEADERS;
    }
}
